/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* string.c
 *
 * implementation of (slightly more) intelligent string library than comes
 * standard in C
 *
 * All functions (except the initialization functions) return zero on success
 * and a positive errno value on failure.
 */

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sisstring.h"

/* For ss_sprintf */
#define INIT_STRSIZE 512
#define MAX_STRSIZE  4096

struct sstring {
        int  allocated;
        int  alias;
        char *end;
        char *start;
};

/**
 * ss_init_sisstring - initialize sisstring based on size
 * @size: size to reserve for string
 *
 * Description: initialize a sisstring of size 'size'.  The given size
 * is the max number of characters that can be stored in the string.
 *
 * Returns: allocated sstring or NULL on error.
 */
struct sstring *ss_init(int size)
{
        void *p;
        struct sstring *s;

        s = p = (void *) malloc(sizeof(*s) + (sizeof(char) * (size-1)));

        if(p != NULL) {
                s->end = NULL;
                /* point start to the extra memory at the end of the struct */
                (void *) s->start = (p + sizeof(*s));
                s->allocated = size;
                s->alias = 0;
        }

        return s;
}

/**
 * ss_init_alias - initialize a sisstring for use as an alias
 * @s: string to init
 *
 * Description: Sets up a sisstring to function as an alias.  That is,
 * it merely points to a string of characters in memory with the symantics of
 * a sisstring.  Saves space and is faster than copying.
 *
 * Returns: allocated sstring or NULL on error.
 */
struct sstring *ss_init_alias(void)
{
        struct sstring *s = malloc(sizeof(*s));

        if(s != NULL) {
                s->alias = 1;
                s->start = NULL;
                s->end = NULL;
                s->allocated = 0;
        }

        return s;
}

/**
 * ss_destroy - destroy an initialized string
 * @s: string to destroy
 *
 * Description: Clears a sisstring and frees any memory contained within it.
 */
int ss_destroy(struct sstring *s)
{
        int error = 0;
        void *p = (void *) s;
        if(s == NULL) {
                error = EFAULT;
        }
        else {
                free(p);
                p = NULL;
        }

        return error;
}

/**
 * ss_copy - copy src to dest
 * @dest: destination string
 * @src: source string
 *
 * Description:  copies src string into dest string.  dest must be at least
 * as big as src string.
 */
int ss_copy(struct sstring *dest, struct sstring *src)
{
        int error = 0;
        int srcsize = 0;
        int destsize = 0;
        /* src need to be fully defined, dest just needs to be initialized */
        if((error = ss_len(src, &srcsize)) != 0)
                goto fail;
        if((error = ss_allocated(dest, &destsize)) != 0)
                goto fail;

        /* we don't allow a copy if the destination sisstring is too small */
        if (destsize < srcsize) {
                error = ERANGE;
                goto fail;
        }

        if(memcpy(dest->start, src->start, srcsize) != 0) {
                error = errno;
                goto fail;
        }
        dest->end = dest->start + srcsize;

 fail:
        return error;
}

/**
 * ss_cat - concatenate two sstrings together
 * @dest: destination string
 * @src: source string
 *
 * Description: concatenates source string to dest string.  Dest must
 * have enough space allocated to contain both what it already
 * contains and the source string.
 *
 */
int ss_cat(struct sstring *dest, struct sstring *src)
{
        int error = 0;
        int srcsize = 0;
        int destsize = 0;
        int destalloc = 0;

        /* src need to be fully defined, dest just needs to be initialized */
        if((error = ss_len(src, &srcsize)) != 0)
                goto fail;
        if((error = ss_len(dest, &destsize)) != 0)
                goto fail;
        if((error = ss_allocated(dest, &destalloc)) != 0)
                goto fail;

        if(srcsize+destsize > destalloc) {
                error = ERANGE;
                goto fail;
        }

        if(memcpy(dest->end+1, src->start, srcsize) != 0) {
                error = errno;
                goto fail;
        }
        dest->end = dest->end+srcsize;

 fail:
        return error;
}

/**
 * ss_cmp - compare two sisstrings
 * @str1: first string
 * @str2: second string
 * @result: location to store result of compare
 *
 * Description: similar to the strcmp fxn.  result is set to less than,
 * equal to, or greater than zero if str1 is less than, equal to, or greater
 * than str2.
 */
int ss_cmp(struct sstring *str1, struct sstring *str2, int *result)
{
        int error = 0;
        int size1 = 0;
        int size2 = 0;
        int i;

        if((error = ss_len(str1, &size1)) != 0)
                goto fail;
        if((error = ss_len(str2, &size2)) != 0)
                goto fail;

        *result = 0;

        for(i = 0; i < size1 && i < size2; i++) {
                if(str1->start[i] < str2->start[i]) {
                        *result = -1;
                        break;
                }
                if(str1->start[i] > str2->start[i]) {
                        *result = 1;
                        break;
                }
        }
        if(*result == 0 && size1 != size2) {
                if(size1 < size2) {
                        *result = -1;
                }
                else {
                        *result = 1;
                }
        }
 fail:
        return error;
}

/**
 * ss_length - returns the length of the given string
 * @s: string to get length of
 * @len: variable to store length in
 *
 * Description: retrieves the length of the sisstring and stores it in the
 * length variable.
 */
int ss_len(struct sstring *s, int *len)
{
        int error = 0;
        *len = 0;
        if(s == NULL || len == NULL)
                error = EFAULT;
        else if (s->start != NULL && s->end != NULL)
                *len = s->end - s->start + 1;

        return error;
}

/**
 * ss_size - returns max size of string
 * @s: string to check
 * @size: variable to store max size in
 *
 * Description: Returns the size that the sisstring was initialized to hold.
 *
 */
int ss_allocated(struct sstring *s, int *array_size)
{
        int error = 0;
        if(s == NULL)
                error = EFAULT;
        else
                *array_size = s->allocated;
        return error;
}

/**
 * ss_alias - create an alias to a sisstring
 * @dest: destination sisstring, initialized as an alias
 * @src: source sisstring, must have content
 * @start: offset from source's start
 * @end: offset from source's start (set to src's end if < 0)
 *
 * Description: Creates an alias to another sisstring.  Start must be less
 * than end, unless end is negative (which means to set the destination
 * end to the source end).  start and end are 0 based.
 */
int ss_alias(struct sstring *dest, struct sstring *src, int start, int end)
{
        int error = 0;
        int size = 0;

        if(end >= 0 && start > end) {
                error = EINVAL;
                goto fail;
        }

        if((error = ss_len(src, &size)) != 0)
                goto fail;

        if(size <= 0)
                error = ERANGE;

        /* FIXME: check is_alias fxn */

        dest->start = src->start + start;
        dest->end = end < 0 ? src->end : src->start + end;
 fail:
        return error;
}

/**
 * cstring2ss_alias - points to the source cstring
 * @dest: sstring to use as an alias
 * @str: source cstring
 * @len: length of source string not including null terminator
 *
 * Description: creates an alias to a standard character string.
 *
 */
int cstring2ss_alias(struct sstring *dest, char *str, int len)
{
        int error = 0;

        /* FIXME: add an is_alias fxn */
        /* FIXME: check of len of zero */

        dest->start = str;
        dest->end = str+len-1;

        return error;
}

/**
 * cstring2ss - copies a cstring to a sstring
 * @dest: sisstring to copy to
 * @src: character string to copy from
 * @len: length of src cstring not including null terminator
 *
 * Description: copies length characters of a standard character string
 * to a sisstring.
 */
int cstring2ss(struct sstring *dest, char *src, int len)
{
        int error = 0;
        int size = 0;
        int i;

        if((error = ss_allocated(dest, &size)) != 0)
                goto fail;
        if(size == 0) {
                error = ERANGE;
                goto fail;
        }
        /* FIXME: Check for len of zero */
        /* we will truncate if the sisstring is smaller than the src */
        if(size > len)
                size = len;

        for(i = 0; i < size; i++) {
                dest->start[i] = src[i];
        }
        dest->end = dest->start+(i-1);

 fail:
        return error;
}

/**
 * ss2cstring - copies a sisstring to a cstring
 * @dest: character pointer to store string
 * @len: total number of chars in dest
 * @src: source sisstring to copy from
 *
 * Description: copies the characters from a sisstring to a standard
 * character string, making sure to null terminate it.  Therefore if
 * length is equal to the source sisstring's length, the last character
 * will be truncated.
 *
 */
int ss2cstring(char *dest, int len, struct sstring *src)
{
        int error = 0;
        int size = 0;
        int i;

        if((error = ss_len(src, &size)) != 0)
                goto fail;
        if(size == 0 || len <= 0) {
                error = ERANGE;
                goto fail;
        }

        /* we will truncate if the dest string is smaller than the src */
        if(size >= len)
                size = len-1;

        for(i = 0; i < size; i++) {
                dest[i] = src->start[i];
        }
        dest[i] = '\0';

 fail:
        return error;
}

/**
 * ss_sprintf - copies a formatted string into a sstring
 * @dest: sstring to copy into
 * @format: format string (ala printf)
 *
 * Description:
 *
 */
int ss_sprintf(struct sstring *dest, char *fmt, ...)
{
        int error = 0;
        int size = 0;
        char *tmpstr = NULL;
        va_list ap;

        va_start(ap, fmt);

        tmpstr = (char *) malloc(sizeof(char) * INIT_STRSIZE);

        if(tmpstr == NULL) {
                error = errno;
                goto fail;
        }

        if((error = vsnprintf(tmpstr, INIT_STRSIZE, fmt, ap)) >= INIT_STRSIZE) {
                char *p;
                size = error+1;
                if(size > MAX_STRSIZE)
                        size = MAX_STRSIZE;

                p = realloc(tmpstr, size);
                if(p == NULL) {
                        error = errno;
                        free(tmpstr);
                        goto fail;
                }
                tmpstr = p;
                if((error = vsnprintf(tmpstr, size, fmt, ap)) >= size) {
                        /* FIXME: valid error return value? */
                        error = EMSGSIZE;
                        free(tmpstr);
                        goto fail;
                }
        }
        size = error;

        if((error = cstring2ss(dest, tmpstr, size)) != 0)
                goto fail;

 fail:
        return error;
}
