/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "gulm_defines.h"
#include "LLi.h"
#include "hash.h"
#include "config_gulm.h"
#include "config_ccs.h"
#include "xdr.h"
#include "utils_ip.h"
#include "utils_dir.h"
#include "utils_crc.h"
#include "utils_verb_flags.h"
#include "ccs.h"

/*****************************************************************************/
/* bits of data used by the log_*() and die() functions. */
extern uint32_t verbosity;
extern char *ProgramName;

/*****************************************************************************/

/**
 * release_node_list - 
 * @list: <> list to free up
 *
 * free up all of the elements in a LLi of ip_name_t.
 * 
 * 
 * Returns: void
 */
void release_node_list(LLi_t *list)
{
   LLi_t *tmp, *nxt;
   ip_name_t *in;
   for(tmp = LLi_next(list);
       NULL != LLi_data(tmp);
       tmp = nxt ) {
      nxt = LLi_next(tmp);
      LLi_del(tmp);
      in = LLi_data(tmp);
      if( in->name != NULL ) free(in->name);
      free(in);
   }
}

/**
 * Can_I_be_a_master - 
 * or, am i in the servers list?
 * Returns: TRUE or FALSE
 */
int Can_I_be_a_master(gulm_config_t *gf, uint32_t ip)
{
   LLi_t *tmp;
   ip_name_t *in;
   if( LLi_empty(&gf->node_list) ) return FALSE;
   for(tmp = LLi_next(&gf->node_list);
       NULL != LLi_data(tmp);
       tmp = LLi_next(tmp)) {
      in = LLi_data(tmp);
      if( in->ip == ip ) {
         return TRUE;
      }
   }
   return FALSE;
}

/*****************************************************************************/

unsigned int calc_quorum(unsigned int quorum, unsigned int cnt)
{
   if( quorum > cnt ) {
      return (cnt / 2 ) + 1;
   }else if( quorum == 0 )
      return 1;
   else
      return quorum;
}


/*****************************************************************************/

/**
 * hash_config - 
 * 
 * generate a hash key for this config.  This is used to make sure all
 * servers are using compatible configs.
 *
 * Note that only the important bits are hashed.  Parts not hashed can
 * differ on nodes, and not affect the cluster's ablity to run.  fencebin
 * is an example, as is the verbosity.
 * 
 * Returns: uint32_t
 */
uint32_t hash_config(gulm_config_t *gf)
{
   register uint32_t hash = 0x6d696b65;
   LLi_t *tmp;
   ip_name_t *in;

   hash = crc32(gf->clusterID, strlen(gf->clusterID), hash);
   hash = crc32((uint8_t*)&gf->corePort, sizeof(uint16_t), hash);
   hash = crc32((uint8_t*)&gf->heartbeat_rate, sizeof(uint64_t), hash);
   hash = crc32((uint8_t*)&gf->allowed_misses, sizeof(uint16_t), hash);
   hash = crc32((uint8_t*)&gf->quorum, sizeof(uint16_t), hash);
   hash = crc32((uint8_t*)&gf->fog, sizeof(uint16_t), hash);

   hash = crc32((uint8_t*)&gf->node_cnt, sizeof(uint16_t), hash);
   for(tmp = LLi_next(&gf->node_list);
       NULL != LLi_data(tmp);
       tmp = LLi_next(tmp) ) {
      in = LLi_data(tmp);
      hash = crc32((uint8_t*)&in->ip, sizeof(uint32_t), hash);
   }

   hash = crc32((uint8_t*)&gf->how_many_lts, sizeof(uint16_t), hash);
   hash = crc32((uint8_t*)&gf->lt_port, sizeof(uint16_t), hash);
   hash = crc32((uint8_t*)&gf->ltpx_port, sizeof(uint16_t), hash);

   return hash;
}

/**
 * dump_conf - 
 * 
 */
void dump_conf(gulm_config_t *gf)
{
   FILE *fp;
   int fd;
   LLi_t *tmp;
   ip_name_t *in;
   uint64_t pf;

   if( (fd=open_tmp_file("Gulm_config")) < 0 ) return;
   if((fp = fdopen(fd,"a")) == NULL ) return;
   fprintf(fp,"#======================================="
         "========================================\n");

   fprintf(fp, "# hashed: %#x\n", gf->hashval);
   fprintf(fp, "cluster {\n name = \"%s\"\n", gf->clusterID);
   fprintf(fp, " lock_gulm {\n");
   pf = gf->heartbeat_rate / 1000;
   fprintf(fp, "  heartbeat_rate = %.3f\n", pf / 1000.0 );
   fprintf(fp, "  allowed_misses = %u\n", gf->allowed_misses);
   fprintf(fp, "  coreport = %u\n", gf->corePort);
   if( gf->quorum <= gf->node_cnt )
      fprintf(fp, "  quorum = %u\n", gf->quorum);
   pf = gf->new_con_timeout / 1000;
   fprintf(fp, "  new_connection_timeout = %.3f\n", pf / 1000.0 );
   fprintf(fp, "  # server cnt: %u\n", gf->node_cnt);

   tmp = LLi_next(&gf->node_list);
   in = LLi_data(tmp);
   fprintf(fp, "  # servers = [\"%s\"", in->name);
   for(tmp = LLi_next(tmp);
       NULL != LLi_data(tmp);
       tmp = LLi_next(tmp) ) {
      in = LLi_data(tmp);
      fprintf(fp, ", \"%s\"", in->name);
   }
   fprintf(fp, "]\n");
   tmp = LLi_next(&gf->node_list);
   in = LLi_data(tmp);
   fprintf(fp, "  servers = [\"%s\"", iptostr(in->ip));
   for(tmp = LLi_next(tmp);
       NULL != LLi_data(tmp);
       tmp = LLi_next(tmp) ) {
      in = LLi_data(tmp);
      fprintf(fp, ", \"%s\"", iptostr(in->ip));
   }
   fprintf(fp, "]\n");

   fprintf(fp, "  lt_partitions = %u\n", gf->how_many_lts);
   fprintf(fp, "  lt_base_port = %u\n", gf->lt_port);
   fprintf(fp, "  lt_high_locks = %lu\n", gf->lt_maxlocks);
   fprintf(fp, "  lt_drop_req_rate = %u\n", gf->lt_cf_rate);
   fprintf(fp, "  prealloc_locks = %u\n", gf->lt_prelocks);
   fprintf(fp, "  prealloc_holders = %u\n", gf->lt_preholds);
   fprintf(fp, "  prealloc_lkrqs = %u\n", gf->lt_prelkrqs);
   fprintf(fp, "  ltpx_port = %u\n", gf->ltpx_port);
   pf = gf->cb_recheck_rate / 1000;
   fprintf(fp, "  cb_recheck_rate = %.3f\n", pf/ 1000.0 ); 
   fprintf(fp, "  \n\n");

   fprintf(fp, " }\n}\n");

   fclose(fp);
}

/**
 * get_lt_range - 
 * @which: 
 * @of:
 * @start: 
 * @stop: 
 * 
 * Given which lt in a set, what would it's start stop range be?
 * For example:  which:3 of:10 => start:78 stop:103
 *
 * Wonder if there is a way to do this with out the loop. probably.
 * 
 */
int get_lt_range(int which, int of, int *start, int *stop)
{
   int i, eg, lo, last=0;

   if( of > 256 ) return -EINVAL;
   if( which > of ) return -EINVAL;

   eg = ( 256 - of ) / of;
   lo = ( 256 - of ) % of;

   for(i = 0 ; i < of ; i++ ) {
      *start = last;
      last += eg;
      if( lo > 0 ) { lo--; last ++; }
      *stop = last;
      last++;

      if( i == which ) break;
   }
   return 0;
}

/**
 * scan_cluster_section - 
 * @gf: 
 * @rt: 
 * 
 * 
 * Returns: int
 */
int scan_cluster_section(gulm_config_t *gf, ccs_node_t *rt)
{
   if(gf->clusterID != NULL ) free(gf->clusterID);
   gf->clusterID = strdup( find_ccs_str(rt, "cluster/name", '/',
                                        "cluster") );
   if( gf->clusterID == NULL ) die(ExitGulm_NoMemory, "Out of Memory.\n");

   if( strlen(gf->clusterID) <= 0 )
      die(ExitGulm_ParseFail, "Cluster name \"%s\" is too short\n",
            gf->clusterID);
   if( strlen(gf->clusterID) > 16 )
      die(ExitGulm_ParseFail, "Cluster name \"%s\" is too long\n",
            gf->clusterID);

   return 0;
}

/**
 * scan_server_list - 
 * @gf: 
 * @nd: 
 * 
 * 
 * Returns: int
 */
int scan_server_list(gulm_config_t *gf, ccs_node_t *nd)
{
   ccs_value_t *val;
   ip_name_t *in;

   if( nd == NULL || nd->v == NULL ) {
      die(ExitGulm_ParseFail,
            "I couldn't find a \"cluster { lock_gulm { servers = [] } }\""
            " section.\n");
   }

   for(val = nd->v; val; val = val->next) {
      if( val->type != CCS_STRING ) { return -1; }
      in = malloc(sizeof(ip_name_t));
      if( in == NULL ) return -1;
      LLi_init( &in->in_list, in);

      if( gulm_aton(val->v.str, &in->ip) == 0 ) {
         char nme[64], *c;
         /* they gave a dotted ip, try to find name */
         /* get things into the expected byte order. */
         in->ip = osi_cpu_to_be32(in->ip);
         if( get_name_for_ip(nme, 64, in->ip) != 0 ) {
            die(ExitGulm_ParseFail, "I cannot find the name for ip \"%s\".\n",
                  val->v.str);
         }else {
            nme[63] = '\0'; /* jic */
            c = strstr(nme, ".");
            if( c != NULL ) *c = '\0';
            in->name = strdup(nme);
            if( in->name == NULL ) die(ExitGulm_NoMemory, "Out of memory.\n");
         }
      }else
      if( load_ip_from_ccs(val->v.str, &in->ip) == 0 ) {
         /* they have usedev */
         in->name = strdup(val->v.str);
         if( in->name == NULL ) die(ExitGulm_NoMemory, "Out of memory.\n");
      }else
      if( get_ip_for_name(val->v.str, &in->ip) == 0 ) {
         /* they gave a name, and we just got the ip */
         in->name = strdup(val->v.str);
         if( in->name == NULL ) die(ExitGulm_NoMemory, "Out of memory.\n");
      }else
      {
         /* not an ip or name of an ip */
         free(in);
         die(ExitGulm_ParseFail, "I cannot find the ip of \"%s\".\n",
               val->v.str);
      }
      if( gf->node_cnt < 5 ) {
         LLi_add_before( &gf->node_list, &in->in_list );
         gf->node_cnt ++;
      }else{
         fprintf(stderr, "Skipping server entry \"%s\" since the max of five"
               " has been reached.\n", val->v.str);
         free(in);
      }

   }

   return 0;
}


/**
 * bound_uint16 - 
 * @val: 
 * @min: 
 * @max: 
 * 
 * given an int, make sure it fits within the range given my the min and
 * max.  and return the value within a uin16_t
 *
 * do the compairs on ints, but return uint16
 *
 * One of the big things I'm trying to do here, is that if given a -1,
 * return the min value.  If we convert to a unsigned first, we'll end up
 * with the max value, and that isn't what we want.
 * 
 * Returns: uint16_t
 */
uint16_t bound_to_uint16(int val, uint16_t min, uint16_t max)
{
   if( val < (int)min ) return min;
   if( val > (int)max ) return max;
   return val;
}

unsigned int bound_to_uint(int val, unsigned int min, unsigned int max)
{
   if( val < min ) return min;
   if( val > max ) return max;
   return val;
}
unsigned long bound_to_ulong(unsigned long val, unsigned long min, unsigned long max)
{
   if( val < min ) return min;
   if( val > max ) return max;
   return val;
}
uint64_t bound_to_uint64(uint64_t val, uint64_t min, uint64_t max)
{
   if( val < min ) return min;
   if( val > max ) return max;
   return val;
}

/**
 * ft2uint64 - 
 * @time: < float in form <sec>.<millisec>
 * 
 * 
 * Returns: uint64_t time in micro-seconds
 */
uint64_t ft2uint64(float time)
{
   return ((uint64_t)(time * 1000)) * 1000;
}

/**
 * find_ulong - 
 * @rt: 
 * @key: 
 * @def: 
 *
 * kludge around ccs libs.
 * 
 * 
 * Returns: long
 */
unsigned long find_ulong(ccs_node_t *rt, char *key, unsigned long def)
{
   ccs_node_t *tmp;
   tmp = find_ccs_node(rt, key, '/');
   if(tmp != NULL && tmp->v != NULL ) {
      switch(tmp->v->type) {
         case CCS_FLOAT:
            return (unsigned long)tmp->v->v.r;
            break;
         case CCS_INT:
            return tmp->v->v.i;
            break;
         case CCS_STRING:
            return strtoul(tmp->v->v.str, NULL, 10);
            break;
         default:
            return def;
            break;
      }
   }
   return def;
}

/**
 * scan_gulm_section - 
 * @gf: 
 * @rt: 
 * 
 * 
 * Returns: int
 */
int scan_gulm_section(gulm_config_t *gf, ccs_node_t *rt)
{
   char *tmpstr;
   float tmp_ft;
#define CGPRE "cluster/lock_gulm/"

   /* find int values */
   gf->corePort = find_ccs_int(rt, CGPRE"coreport", '/', 40040);
   gf->ltpx_port = find_ccs_int(rt, CGPRE"ltpx_port", '/', 40042);
   gf->lt_port = find_ccs_int(rt, CGPRE"lt_base_port", '/', 41040);

   tmp_ft = find_ccs_float(rt, CGPRE"heartbeat_rate", '/', 
         find_ccs_int(rt, CGPRE"heartbeat_rate", '/',15.0));
   gf->heartbeat_rate = bound_to_uint64(ft2uint64(tmp_ft), 75000,(uint64_t)~0);

   gf->allowed_misses = bound_to_uint16( find_ccs_int(rt,
            CGPRE"allowed_misses", '/', 2), 1, 0xffff);

   tmp_ft = find_ccs_float(rt, CGPRE"new_connection_timeout", '/',
         find_ccs_int(rt, CGPRE"new_connection_timeout", '/',15.0));
   gf->new_con_timeout = bound_to_uint64(ft2uint64(tmp_ft), 0,(uint64_t)~0);

   tmp_ft = find_ccs_float(rt, CGPRE"master_scan_delay", '/',
         find_ccs_int(rt, CGPRE"master_scan_delay", '/',1.0));
   gf->master_scan_delay = bound_to_uint64(ft2uint64(tmp_ft), 10, (uint64_t)~0);

   gf->how_many_lts = bound_to_uint16( find_ccs_int(rt,
            CGPRE"lt_partitions", '/', 1), 1, 256);

   gf->lt_maxlocks = bound_to_ulong(find_ulong(rt, CGPRE"lt_high_locks",
            1024*1024), 10000, ~0UL);

   gf->lt_prelocks = bound_to_uint( find_ccs_int(rt,
            CGPRE"prealloc_locks", '/', 90000 ), 0, ~0U);
   gf->lt_preholds = bound_to_uint( find_ccs_int(rt,
            CGPRE"prealloc_holders", '/', 130000 ), 0, ~0U);
   gf->lt_prelkrqs = bound_to_uint( find_ccs_int(rt,
            CGPRE"prealloc_lkrqs", '/', 60 ), 0, ~0U);

   gf->lt_cf_rate = bound_to_uint( find_ccs_int(rt,
            CGPRE"lt_drop_req_rate", '/', 10), 5, ~0U);

   tmp_ft = find_ccs_float(rt, CGPRE"cb_recheck_rate", '/', 
         find_ccs_int(rt, CGPRE"cb_recheck_rate", '/',15.0));
   gf->cb_recheck_rate = bound_to_uint64(ft2uint64(tmp_ft), 0,(uint64_t)~0);

   /* find strings that are copied in. */
   if(gf->fencebin != NULL ) free(gf->fencebin);
   gf->fencebin = strdup( find_ccs_str(rt,
            CGPRE"fencebin", '/', "fence_node") );
   if( gf->fencebin == NULL ) die(ExitGulm_NoMemory, "Out of Memory.\n");

   if(gf->run_as != NULL ) free(gf->run_as);
   gf->run_as = strdup( find_ccs_str(rt, CGPRE"run_as", '/', "root") );
   if( gf->run_as == NULL ) die(ExitGulm_NoMemory, "Out of Memory.\n");

   if(gf->lock_file != NULL ) free(gf->lock_file);
   gf->lock_file = strdup( find_ccs_str(rt, CGPRE"lock_dir", '/',
                                        "/var/run/sistina") );
   if( gf->lock_file == NULL ) die(ExitGulm_NoMemory, "Out of Memory.\n");

   if(find_ccs_str(rt, CGPRE"lock_file", '/', NULL ) != NULL ) {
      fprintf(stderr, "Warning! "CGPRE"lock_file is depercated, use "
            CGPRE"lock_dir instead.\n");
   }

   /* find strings that are parsed more */
   tmpstr = (char*)find_ccs_str(rt, CGPRE"verbosity", '/', NULL);
   if( tmpstr != NULL ) set_verbosity(tmpstr, &verbosity);

   /* now get the list of servers. */
   if( scan_server_list(gf, find_ccs_node(rt, CGPRE"servers", '/')) < 0 )
      die(ExitGulm_ParseFail, "Failed to parse all of the servers list.\n");

#undef CGPRE

   /* no one should ever change these, so we won't make them changeable */
   gf->node_verification_type = nvt_allow;
   gf->lt_hashbuckets = 65536;
   gf->quorum = 0xfffe;

   return 0;
}

/**
 * rebuild_server_list - 
 * @gf: 
 *
 * this cannot be done on a live cluster if in SLM.
 * 
 * flushes old server list.
 * re-reads it from ccs.
 * re-calculates config crc.
 * 
 * Returns: int
 */
int rebuild_server_list(gulm_config_t *gf)
{
   int err = 0;
   ccs_node_t *rt;

   if( ! gf->fog ) return -EINVAL;

   if((err=open_ccs_file(&rt, "cluster.ccs")) != 0 ) {
      log_msg(lgm_Always, "Failed to get \"cluster.ccs\" from ccsd: %d:%d:%s\n",
            err, errno, strerror(errno));
      return err;
   }

   /* remove old list */
   release_node_list(&gf->node_list);
   gf->node_cnt = 0;
   LLi_init_head(&gf->node_list);

   err = scan_server_list(gf,find_ccs_node(rt,"cluster/lock_gulm/servers",'/'));

   /* calc hash. */
   gf->hashval = hash_config(gf);

   close_ccs_file(rt);
   return err;
}

/**
 * parse_conf - 
 * @gf: 
 *
 * Returns: int
 */
int parse_conf(gulm_config_t *gf)
{
   ccs_node_t *rt;
   int err=0;

   LLi_init_head( &gf->node_list );

   if((err=open_ccs_file(&rt, "cluster.ccs")) != 0 ) {
      log_msg(lgm_Always, "Failed to get \"cluster.ccs\" from ccsd: %d:%d:%s\n",
            err, errno, strerror(errno));
      return err;
   }

   /* get cluster section info */
   if( (err = scan_cluster_section(gf, rt)) < 0 ) {
      goto exit;
   }

   /* get cluster/lock_gulm info */
   if( (err = scan_gulm_section(gf, rt)) < 0 ) {
      goto exit;
   }

   /* more than one server? then we must be foggy. */
   gf->fog = (gf->node_cnt > 1);

   /* calc hash. */
   gf->hashval = hash_config(gf);

   /* calc quorum */
   gf->use_quorum = calc_quorum(gf->quorum, gf->node_cnt);
   if( gf->fog == FALSE ) gf->use_quorum = 1;

   /* all done. */
exit:
   close_ccs_file(rt);
   return err;
}

/**
 * get_ip_for_name_ccs_libc -
 * check for name in ccs before passing it on to libc
 * @name:
 * @ip: > ip in BigEn.
 *
 * Returns: int
 */
int get_ip_for_name_ccs_libc (char *name, uint32_t * ip)
{
   int err;

   err = load_ip_from_ccs (name, ip);
   if (err)
      return get_ip_for_name_libc (name, ip);
   return 0;
}
/* vim: set ai cin et sw=3 ts=3 : */
